/* --------------------------------------------------------------------
 * Clac - calculator (C) 1994-2006 by Eero Tamminen.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms specified in the COPYRIGHTS file.
 *
 *  Header for Clac, Conv and Front.
 * --------------------------------------------------------------------
 */

#ifndef CLAC_H
#define CLAC_H

#include <math.h>                /* constants & functions  */
#include "complex.h"

/* define some preprosessor symbols          */
#define VARS     32              /* number of variables          */
#define MAX_VAR  8               /* max. variable name lenght    */
#define VAR_NAM  VARS * MAX_VAR  /* stack for variable names     */
#define MAX_LN   256             /* max length for an expression */

/* define error codes */
#define PRG_ERR  88      /* Program error */
#define NO_EXP   1       /* no expression given         */
#define GEN_ERR  2       /* error in expression         */
#define PAR_ERR  3       /* mismatched parenthesis      */
#define DEF_ERR  4       /* undefined result 1/0,lg(-1) */
#define MOD_ERR  5       /* unrecogniced mode           */
#define OVR_ERR  6       /* mode overflow               */
#define VAR_ERR  7       /* variable not defined        */
#define FNC_ERR  8       /* unrecogniced function       */
#define STK_ERR  9       /* par/val/op stack full       */
#define ARG_ERR  10      /* wrong number of arguments   */
#define CPX_ERR  11      /* complex calculation error   */
#define OVF_ERR  12      /* Overflow.                   */

/* define mode identifiers (same as in popup's)                  */
enum type_out { BIN = 1, OCT, DEC, HEX };  /* output modes       */
enum type_trg { RAD = 1, DEG };            /* trig. / cplx modes */

/* define which character indicates which type of number on expression  */
#define BIN_SYM '\''                            /* binary decimal       */
#define OCT_SYM '#'                             /* octal decimal        */
#define HEX_SYM '$'                             /* hexadecimal          */

#ifndef TRUE
#define FALSE  0
#define TRUE  !FALSE
#endif

/* communication structure between clac and the main application  */
struct packet {
  int error;                /* evaluation error             */
  int err_pos;              /* pos. of error in exp. string */
  int mode_trig;            /* trigonometric mode           */
  cplx result;              /* result of evaluation         */
  char *exp;                /* expression to calculate      */
  char vname[VAR_NAM + 2];  /* names of the variables       */
  cplx value[VARS + 1];     /* values of the variables      */
};

/* Exp[] and trig_mode should be filled when calc() is called.
 * Optionally also vname[] and value[] arrays can be filled up.
 * Variable names have to be terminated with a NULL.
 * Calc() returns answer in result + a possible error in error.
 */

#ifdef CLAC_C
#define ext
#else
#define ext extern
#endif

ext void calc(struct packet *pkt);    /* evaluate expression  */

#undef ext

#endif /* CLAC_H */
