/* ----------------------------------------------------------- 
 * Clac - calculator (C) 1994-2006 by Eero Tamminen.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms specified in the COPYRIGHTS file.
 * 
 * common userinterface functions for clac
 */

#include <ctype.h>    /* toupper() */
#include <stdio.h>    /* sprintf() */
#include <string.h>   /* strcpy()  */

#define COMMON_C
#include "clac.h"
#include "common.h"
#undef  COMMON_C

/* --------------------------------------------------------------------  */
/* get a variable name */
char *var_get(int i, char name[], char exp[], int *err)
{
  int a = 0, b = 0;
  char mark;

  /* search for variable name (same criteria as in clac) */
  while(exp[a] <= ' ') a ++;

  mark = toupper (exp[a]);
  if(i > a && ((mark >= 'A' && mark <= 'Z') || (mark >= '0' && mark <= '9')))
  {
    /* store variable name (ignore erranous end) */
    do
    {
      name[b ++] = mark;
      mark = toupper (exp[b + a]);
    } while(((mark >= 'A' && mark <= 'Z') || (mark >= '0' && mark <= '9')) && b < MAX_VAR);
    name[b] = '\0';

    while(mark != '=')
    {
      if(mark != ' ')
      {
        *err = GEN_ERR;    /* syntax error  */
        return (exp);
      }
      mark = exp[++b + a];
    }
    return (&exp[i + 1]);
  }
  else
    *err = GEN_ERR;        /* syntax error  */
  return(exp);
}

/* put variable into stack */
void var_put(char name[], struct packet *comm)
{
  int i = 0, lenny, idx = 0;

  /* find the index for the given variable */
  lenny = strlen(&comm->vname[i]);
  while((lenny > 0) && (strcmp (&comm->vname[i], name)))
  {
    idx ++;
    i += ++lenny;
    lenny = strlen(&comm->vname[i]);
  }

  /* if name fits  */
  if((lenny + i < VAR_NAM) && (idx < VARS))
  {
    strcpy(&comm->vname[i], name);
    comm->value[idx] = comm->result;
  }
  else
    comm->error = STK_ERR;
}

/* error -> error text  */
void errors(int error, char *err)
{
  char *msg;

  switch (error)
  {
    case GEN_ERR:
      msg = "Syntax error.";
      break;
    case PAR_ERR:
      msg = "Mismatched parenthesis.";
      break;
    case DEF_ERR:
    case ERR_DERO:
      msg = "Undefined.";
      break;
    case MOD_ERR:
      msg = "Mode error.";
      break;
    case OVR_ERR:
      msg = "Mode overflow.";
      break;
    case OVF_ERR:
      msg = "Overflow.";
      break;
    case VAR_ERR:
      msg = "Variable undefined.";
      break;
    case FNC_ERR:
      msg = "Unrecognized function.";
      break;
    case STK_ERR:
      msg = "Stack(s) full.";
      break;
    case ARG_ERR:
      msg = "Wrong number of arguments.";
      break;
    case NO_EXP:        /* no expression  */
      msg = "No expression.";
      break;
    case ERR_TYPE:        /* complex errors  */
      msg = "Complex number type error.";
      break;
    default:
      msg = "Program error -> report a bug!";
  }
  strcpy(err, msg);
}

/* --------------------------------------------------------------------
 *               output according to the mode
 * --------------------------------------------------------------------
 */

/* a value, string for ascii string, complex & trig. output modes  */
int conv(cplx value, char *exp, int out, int cpx, int trg)
{
int error = 0;

/* a real number */
  if(value.imag == 0.0)
  {
    switch(out)
    {
      case DEC:
        sprintf(exp, "%g", value.real);
        break;
      case BIN:
        error = make_out(exp, BIN_SYM, 1, value.real);
        break;
      case OCT:
        error = make_out(exp, OCT_SYM, 3, value.real);
        break;
      case HEX:
        error = make_out(exp, HEX_SYM, 4, value.real);
        break;
      default:
        error = MOD_ERR;
    }
  }
  else        /* a complex number  */
  {
    switch(cpx)
    {
      case CPX_CMPNT:
        sprintf(exp, "%g + %gi", value.real, value.imag);
        break;
      case VEC_CMPNT:
        sprintf(exp, "(%g, %g)", value.real, value.imag);
        break;
      case VEC_POLAR:
        value = change_type(value, POLAR);
        if(trg == DEG)
          sprintf(exp, "(%g, %gdeg)", value.real, TO_DEG(value.imag));
        else
          sprintf(exp, "(%g, %grad)", value.real, value.imag);
        break;
      default:
        error = MOD_ERR;
    }
  }
  return(error);
}

/* --------------------------------------------------------------------
 * BIN, OCT and HEX outputs
 */

/* output in 2^bits modes */
int make_out(char exp[], char mode, int bits, double value)
{
  int i, index, mask, num, exp_idx = 0;
  unsigned long num1, num2;          /* int/decimal  parts   */
  int len_long = sizeof(long) * 8;   /* bits in 'long'       */
  char *base = "0123456789ABCDEF";   /* number base(s)       */
  if(bits == 1) len_long --;         /* eliminate negate     */
  mask = (1 << bits) - 1;            /* mask for basedigit   */
  if(len_long % bits)                /* for uneven bitamount */
    len_long -= len_long % bits;

  if(value > ((1L << (len_long - 1)) - 1))
    return(OVR_ERR);
  else
  {
    /* sign      */
    if(value < 0)
    {
      exp[exp_idx ++] = '-';
      value = fabs(value);
    }
    exp[exp_idx ++] = mode;    /* symbol    */
    num1 = value;
    num2 = (value - num1) * (1L << (len_long - bits));

    /* integer part */
    if(num1)
    {
      num = 0;
      i = len_long - bits;
      while(i >= 0)
      {
        index = (num1 >> i) & mask;
        num |= index;
        if(num)
          exp[exp_idx ++] = base[index];
        i -=bits;
      }
    }
    /* zero before point  */
    else
      exp[exp_idx ++] = '0';

    /* decimal part */
    if(num2)
    {
      num = 0;
      exp[exp_idx ++] = '.';
      i = len_long - bits;
      while(i >= 0)
      {
        i -=bits;
        index = (num2 >> i) & mask;
        exp[exp_idx ++] = base[index];
      }
    }
  }
  /* end output string  */
  exp[exp_idx] = '\0';

  return(0);
}
