/* --------------------------------------------------------------------
 * 
 * Clac - calculator (C) 1994-2006 by Eero Tamminen.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms specified in the COPYRIGHTS file.
 * 
 * Code for complex number calculations
 * --------------------------------------------------------------------
 */

#define COMPLEX_C
#include <math.h>
#include "complex.h"
#undef  COMPLEX_C

int cplx_error;

/* ------------------ calculating functions --------------------------- */

/* convert a double to a complex number */
cplx to_cplx (double x)
{
  cplx z;

  z.real = x;
  z.imag = 0.0;

  return (z);
}

/* absolute value (vector lenght) */
double c_abs (cplx x)
{
  return (sqrt ( x.real * x.real + x.imag * x.imag));
}

/* compare two complex numbers -> -1, 0 or 1 */
int c_cmp (cplx x, cplx y)
{
  double a, b;

  a = x.real * x.real + x.imag + x.imag;
  b = y.real * y.real + y.imag + y.imag;

  if (a > b)
    return (1);
  else if (a == b)
    return (0);
  else
    return (-1);
}

/* add x and y together */
cplx cplx_add (cplx x, cplx y)
{
  cplx z;

  z.real = x.real + y.real;
  z.imag = x.imag + y.imag;

  return (z);
}

/* subtract y from x */
cplx cplx_sub (cplx x, cplx y)
{
  cplx z;

  z.real = x.real - y.real;
  z.imag = x.imag - y.imag;

  return (z);
}

/* multiply x and y */
cplx cplx_mul (cplx x, cplx y)
{
  cplx z;

  z.real = x.real * y.real - x.imag * y.imag;
  z.imag = x.real * y.imag + y.real * x.imag;

  return (z);
}

/* divide x with y */
cplx cplx_div (cplx x, cplx y)
{
  double s;
  cplx z;

  s = y.real * y.real + y.imag * y.imag;
  if (s == 0.0)
    cplx_error = ERR_DERO;
  else
  {
    z.real = (x.real * y.real + x.imag * y.imag) / s;
    z.imag = (y.real * x.imag - x.real * y.imag) / s;
  }

  return (z);
}

/* raise x to the power of y (crude) */
cplx cplx_pow (cplx x, double y)
{
  cplx z;

  z = change_type (x, POLAR);
  z.real = pow(z.real, y);
  z.imag = y * z.imag;
  z = change_type (z, RECT);

  return (z);
}

/* translate between component and pointer representations */
cplx change_type (cplx x, int to_type)
{
  cplx z;

  switch (to_type)
  {
    case POLAR:
      /* change to polar: lenght & heading (in radians) */
      z.real = sqrt (x.real * x.real + x.imag * x.imag);
      if (x.real == 0.0)
        if (x.imag == 0.0)
          z.imag = 0.0;
        else
          z.imag = (x.imag < 0.0 ? -1 : 1) * M_PI_2;
      else
        z.imag = atan (x.imag / x.real);
      break;
    case RECT:
      /* change pointer to rectangular representation */
      z.real = x.real * cos (x.imag);
      z.imag = x.real * sin (x.imag);
      break;
    default:
      cplx_error = ERR_TYPE;
  }
  return (z);
}
