/* A barebones filter/list/client-server version of Clac.
 *
 * Calculation engine (C) 1994 by Eero Tamminen.
 * (puujalka@modeemi.cs.tut.fi)
 *
 * Client-server part (C) 1994 by Jarkko Kniivil"a.
 * (jkn@modeemi.cs.tut.fi)
 *
 * Freeware (or under GPL terms).
 */

#include <stdio.h>				/* file stuff		*/
#include <string.h>				/* strcpy/cmp/len()	*/
#include "bare.h"

/* -------------------------------------------------------------------- */
/* define some preprosessor symbols					*/

#define MAX_LN	256			/* max length for an expression */

/* -------------------------------------------------------------------- */
/* declare cli subfunctions						*/

int	parse(int argc, char *argv[], FILE **fp, char **oper, int *len);
int	evaluate(FILE *fp, char *oper, int len, double *total);
char	*errors (int error);
void	help(void);


/* ==================================================================== */
/*				main					*/
/* ==================================================================== */

int main(int argc, char *argv[])
{
	FILE *fp;
	char *oper;
	double total = 0.0;
	int len, error = FALSE;

	/* parse the commandline					*/
	if (parse(argc, argv, &fp, &oper, &len))
	{
		error = evaluate(fp, oper, len, &total);
		fclose(fp);
	}
	else
		error = CMD_ERR;

	if (error)
	{
		printf("%s\n", errors(error));
		if (error == CMD_ERR)
			help();
	}
	else
		printf("%g\n", total);

	return (error);
}

/* -------------------------------------------------------------------- */
/*			parse a commandline				*/
/* -------------------------------------------------------------------- */

int parse(int argc, char *argv[], FILE **fp, char **oper, int *len)
{
	/* filename and possibly a list operation */
	if (argc < 2 || argc > 3)
		return (FALSE);

	if (strcmp(argv[1], "-"))
	{
		if (!(*fp = fopen(argv[1], "r")))
			return (FALSE);
	}
	else
		*fp = stdin;

	if (argc == 3)
	{
		*len = strlen(argv[2]);
		/* cut operation into substrings */
		for(*oper = argv[2] + *len; *oper >= argv[2]; (*oper) --)
			if (**oper == 'x' || **oper == 'X')
				**oper = '\0';
		*oper = argv[2];
	}
	else
	{
		*len = 0;
		*oper = "";
	}
	return (TRUE);
}

/* -------------------------------------------------------------------- */
/*		Evaluate list expressions and operate on them		*/
/* -------------------------------------------------------------------- */
int evaluate(FILE *fp, char *oper, int len, double *total)
{
	char in[MAX_LN], val[32];
	static struct clac_out out = {0.0, 0};
 	int oper_i;

	while (fgets(in, MAX_LN, fp))
	{
		/* calculate an expression from the input list */
		if (calc(in, &out))
			return (out.error);
		else
		{
			/* compose a new expression for the operation */
			if (len)
			{
				strcpy(in, oper);
				oper_i = strlen(oper) + 1;
				while (oper_i <= len)
				{
					sprintf(val, "%g", out.result);
					strcat(in, val);
					strcat(in, &oper[oper_i]);
					oper_i += strlen(oper) + 1;
				}
				/* evaluate expression operation */
				if (calc(in, &out))
					return (out.error);
			}
			*total += out.result;
		}
	}
	return (out.error);
}

char *errors (int error)		/* error -> error text	*/
{
	char *err;

	switch (error) {
	case CMD_ERR:
		err = "Command line error.";
		break;
	case GEN_ERR:
		err = "Syntax error.";
		break;
	case PAR_ERR:
		err = "Mismatched parenthesis.";
		break;
	case DEF_ERR:
		err = "Undefined.";
		break;
	case STK_ERR:
		err = "Stack(s) full.";
		break;
	default:
		err = "Unrecognized error.";		/* :-)	*/
	}
	return (err);
}

void help()
{
	puts("Clac \'bare bones\' v 0.9 (C) 1994 by Eero Tamminen.");
	puts("");
	puts("Synopsis: clac file [expression-list-operation].");
	puts("");
	puts("Replacing filename with a \'-\' indicates that accessing");
	puts("should be done from stdin instead of a file.");
	puts("");
	puts("Optional operation argument indicates what has to be done");
	puts("to the newline separated expressions before they are summed.");
	puts("X in the operation will be replaced with the expression.");
	puts("");
	puts("Expressions may include the operators: +, -, *, /, ^, ( and ).");
}
