/* ------------------------------------------------------------- 
 * Clac - calculator (C) 1994-2006 by Eero Tamminen.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms specified in the COPYRIGHTS file.
 */

/* My first ever GEM program in (SozobonX C v2) C.
 * Compile with 'make -f makegui.mak' when using make coming with
 * the sozobonX v2.
 *
 * Many thanks for the original Sozobon crew,
 * Ian Leopore (GEM handling is mostly from his GEMFast examples)
 * and especially to Jerry Geiger :-).
 *
 * 1.8.1994, Eero Tamminen
 */

/* ------------------------------------ */
#include <stdio.h>		/* FILE etc.    */
#include <ctype.h>		/* toupper()    */
#include <string.h>		/* memcpy(), strcpy() etc. */
#include <osbind.h>		/* Pterm() etc  */
#include <limits.h>		/* PATHSIZE */
#include <signal.h>		/* signal IDs */
#include <vaproto.h>		/* VA-protocol  */
#include <xgemfast.h>		/* AES / VDI	*/

#include "clac.h"		/* this first as it includes complex def. */
#include "common.h"
#include <macros.h>		/* min(), max() */

/* resource object indexes */
#include "clac_rsc.h"
#define RSCFILE	"clac_rsc.rsc"
#define HELPFILE "*:\\CLAC.HYP"
#define HELP_PRG "ST-GUIDE"

#ifndef FALSE
#define FALSE 0
#define TRUE  !FALSE
#endif

/* WINX window bottomed message? */
#define WM_BOTTOMED 33

/* Some key codes: */
#define DEL	0x7F
#define TAB	0x09
#define HELP	0x6200

#define MY_MU		(MU_MESAG | MU_BUTTON | MU_KEYBD)
#define WIN_KIND	(CLOSER | NAME | MOVER)
#define NO_WINDOW	-1

/* define some alerts box texts */
#define ALERT_RSC   "[1][RSC file not found! | |Please locate...][ OK ]"
#define ALERT_HELP  "[1][Can't show help as |ST-Guide isn't available! |][ OK ]"
#define ALERT_CLIP  "[1][|SCRAP.TXT not found! |][  OK  ]"
#define ALERT_SCRAP "[1][|Unable to save |SCRAP.TXT! |][ OK ]"
#define ALERT_CLEAR "[1][|Clear variable |names & values? ][OK|Cancel]"
#define ALERT_WIN   "[1][|Can't open |a new window! |][ Fatal ]"
#define ALERT_FATAL "[1][|Can't load a RSC! |][ Fatal ]"
/* define some fileselector info lines */
#define FSEL_RSC    "Locate RSC..."
#define FSEL_OPEN   "Load expressions..."
#define FSEL_SAVE   "Save expressions..."

/* ------------------------------------ */
/* global variables */
static OBJECT *tree_menu;
static OBJECT *tree_wind;
static OBJECT *tree_mode;
static OBJECT *tree_cplx;
static OBJECT *tree_trig;
static int     win_handle = NO_WINDOW;
static char    win_name[] = "  Clac v1.1  ";
static GRECT   win_rect;	/* window work area  */
static GRECT   gl_rwdesk;	/* desktop work area */

static int menu_id;

static char filename[33]  = RSCFILE;
static char filepath[256];

/* characters corresponding to dialog buttons / shortcuts */
static char ex[64];

static int  mode_cplx = CPX_CMPNT;	/* cpx */
static int  mode_trig = RAD;		/* rad */
static int  mode_out  = DEC;		/* dec */

/* communication structure
 * error, err_pos, mode_trig, result, expression, var. names & values
 */
struct packet comm = {FALSE, 0, RAD, {0.0, 0.0}, "",
                     "F0\0F1\0F2\0F3\0F4\0F5\0F6\0F7\0F8\0F9\0PI\0E\0",
                     {{0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},
                      {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0}, {0.0, 0.0},
                      {0.0, 0.0}, {0.0, 0.0}, {M_PI, 0.0}, {M_E, 0.0}}};


#define HISTORY 16	/* how many history entries */

/* MAX_LN = max. lenght of an expression */
#define SHOWLEN 35	/* lenght of expression string in the dialog */
#define MESSLEN 31
struct {
  int  len;
  int  pos;
  int  idx;
  int  charlen;
  char  exp[MAX_LN + 1];
  char  show[SHOWLEN + 1];
  char  message[MESSLEN + 1];
  char  result[2 * sizeof (long) * 8 + 2];   /* <8*binary>.<8*binary>, < MAX_LN */
} show = {0, 0, 0, 0, "", "", "", "0"};

/* a history ringbuffer for expressions */
struct ring {
  int explen;
  char exp[MAX_LN + 1];
  struct ring *prev;
  struct ring *next;
} buffer[HISTORY];

static struct ring *history, *cur_hist;

/* ------------------------------------
 * Function prototypes.
 */
void	initialize ();

/* GEM handling functions. */
void    prg_init (void);
int     w_open (void);
void    msg_handler (short *msgbuf);
int     mouse_handler (int mx, int my, int mb);
int     key_handler (char key);
void    rb_select (OBJECT *tree, int selobj);
int     key_to_obj (OBJECT *tree, char key);
void    do_redraw (short handle, OBJECT *tree, GRECT *redraw_rect);
void	de_select (OBJECT *ptree, int selobj);
int     popup (OBJECT *popup, int defobj, OBJECT *parent, int selobj);
int	do_dialog (int rsc_tree);
void    w_close (void);
void    prg_exit (void);
int	select_file (char *mask, char *filename, char *message);
char   *pathcut (char *filepath);

/* Selection handling. */

/* Menu innards... */
void	do_menu (short entry);
char	*scrap_init(void);
void	read_clip(void);
void	write_clip(char *string);
void	clear_vars ();
void	open_file (int mode);		/* load and save expressions */

/* Dialog innards... */
void    do_object (int object, short key, short state);
void	memory (int memory, char *string, int state);
void	change_mode (OBJECT *ptree, int *mode, int selobj);
void	handle_history (int direction);
void	add_char (char key);
void	del_char (void);
void	show_string (void);
void	evaluate (void);


/* ------------------------------------
 * Standard GEM handling functions (main loop).
 */

void main()
{
  short  mbuf[8];
  short  evnt_m, m1fl, pmx1, pmy1, pm1w, pm1h,
         m2fl, pmx2, pmy2, pm2w, pm2h, pmx, pmy, pmstate,
         state, key, clicks, tlo_cnt, thi_cnt;
  int  selobj;
  m1fl = pmx1 = pmy1 = pm1w = pm1h = 0;
  m2fl = pmx2 = pmy2 = pm2w = pm2h = 0;
  tlo_cnt = thi_cnt = 0;

  prg_init ();
  initialize ();

  if (_app)
  {
    graf_mouse (ARROW, 0);
    w_open ();
  }

  while (TRUE)
    {
      /* Events: mouse button, mouse co-ordinate 1,
       * mouse co-ordinate 2, message buffer (valid address!!!), timer.
       * Mouse co-ordinate, mouse button, special key and normal key states.
       * Number of mouse clicks.
       */
      evnt_m = evnt_multi(MY_MU, 259, 3, 0, m1fl, pmx1, pmy1, pm1w, pm1h,
               m2fl, pmx2, pmy2, pm2w, pm2h, mbuf, tlo_cnt, thi_cnt,
               &pmx, &pmy, &pmstate ,&state, &key, &clicks);

      if (evnt_m & MU_MESAG)
        {
          /* handle messages (menu, window gadgets, redraw...) */
          msg_handler (mbuf);
        }

      selobj = NO_OBJECT;

      if (evnt_m & MU_KEYBD)
        {
          if (state & K_CTRL)
            {
              /* normal key to upper case */
              if (key & 0xFF)
                key = (key & 0xFF) | 0x40;

              /* if a menu shortcut (control + a key) */
              do_menu (key);
            }
          else
            {
              /* or a dialog shortcut key */
              if (key & 0xFF)
                {
                  /* normal, ascii keys to upper keys */
                  selobj = key_handler (toupper (key & 0xFF));
                  /* not shortcut, just a key */
                  if (selobj == NO_OBJECT)
                    {
                      key = key & 0xFF;
                      selobj = 0;
                    }
                }
              else
                {
                  if (key == HELP)
                    do_menu (HELP);
                  else
                    /* map special keys to lower letters */
                    selobj = key_handler (((key >> 8) + 40) & 0xFF);
                }

            }
        }

      if (evnt_m & MU_BUTTON)
        {
          /* mouse: co-ordinates, buttons */
          selobj = mouse_handler (pmx, pmy, pmstate);
        }

      if (selobj != NO_OBJECT)
        {
          de_select (tree_wind, selobj);
          do_object (selobj, key, state);
        }

    } /* Endless loop. */
}

/* ------------------------------------
 * program specific initalization
 */
void initialize (void)
{
  int i;

  /* relocate text strings from rsc to program variables */
  tree_wind[RESULT].ob_spec.tedinfo->te_ptext = show.result;
  tree_wind[MESSAGES].ob_spec.tedinfo->te_ptext = show.message;
  tree_wind[EXPRESS].ob_spec.tedinfo->te_ptext = show.show;

  show.charlen = tree_wind[EXPRESS].ob_width / SHOWLEN;
  show.show[SHOWLEN] = '\0';

  /* set history pointers */
  for (i = 0; i < HISTORY; i ++)
    {
      buffer[i].explen = 0;
      strcpy (buffer[i].exp, "\0");
      buffer[i].prev = &buffer[i - 1];
      buffer[i].next = &buffer[i + 1];
    }
  buffer[0].prev = &buffer[HISTORY - 1];
  buffer[HISTORY - 1].next = buffer;
  cur_hist = history = buffer;

  ex[POINT] = '.';
  ex[ZERO] = '0';
  ex[ONE] = '1';
  ex[TWO] = '2';
  ex[THREE] = '3';
  ex[FOUR] = '4';
  ex[FIVE] = '5';
  ex[SIX] = '6';
  ex[SEVEN] = '7';
  ex[EIGHT] = '8';
  ex[NINE] = '9';
  ex[MULTIPLY] = '*';
  ex[DIVIDE] = '/';
  ex[PLUS] = '+';
  ex[MINUS] = '-';
  ex[TOPOWER] = '^';
  ex[PARLEFT] = '(';
  ex[PARIGHT] = ')';

  Dgetpath (filepath, 0);
  strcat (filepath, "\\*.*");
  filename[0] = '\0';
}

/* ------------------------------------
 * Standard GEM startup.
 */
void prg_init (void)
{
  gl_apid = appl_init ();

  if (!rsrc_load(filename))
    {
      if (!_app)
        {
          /* I'm an acc and have nothing to do... */
          form_alert (1, ALERT_FATAL);
          while (TRUE) evnt_timer (-1, -1);
        }

      form_alert (1, ALERT_RSC);
      /* get current path */
      Dgetpath (filepath, 0);
      strcat (filepath, "\\*.RSC");
      if (select_file (filepath, filename, FSEL_RSC))
        {
          strcat (pathcut (filepath), filename);
          if (!rsrc_load(filepath))
            {
              form_alert (1, ALERT_FATAL);
              appl_exit ();
              Pterm (1);
            }
        }
      else
        {
          appl_exit ();
          Pterm (1);
        }
    }

  /* get desktop work area */
  winx_get (0, WF_WORKXYWH, &gl_rwdesk);


  /*  POPMENU etc. should be defines, generated by the rsc editor */
  rsrc_gaddr (R_TREE, POPMENU, &tree_menu);
  rsrc_gaddr (R_TREE, TREEMAIN, &tree_wind);
  rsrc_gaddr (R_TREE, POPUPDEC, &tree_mode);
  rsrc_gaddr (R_TREE, POPUPCPX, &tree_cplx);
  rsrc_gaddr (R_TREE, POPUPRAD, &tree_trig);

  if (!_app)
    menu_id = menu_register (gl_apid, win_name);	/* register acc */

  /* MiNT: on a case of an exiting request */
  signal(SIGHUP, prg_exit);
  signal(SIGINT, prg_exit);
  signal(SIGQUIT, prg_exit);
  signal(SIGTERM, prg_exit);
}

/* Standard GEM exit. */
void prg_exit (void)
{
  if (!_app)
    {
      /* accs don't quit */
      w_close ();
    }
  else
    {
      w_close ();
      rsrc_free ();
      appl_exit ();
      Pterm (0);
    }
}

/* fileselector */
int select_file (char *mask, char *filename, char *message)
{
  short  button;

  return (fsel_exinput (mask, filename, &button, message) && button);
}

/* Create and open a window. */
int w_open (void)
{
  GRECT  wrect;
  GRECT  trect;

  if (win_handle != NO_WINDOW)          /* if window is already open,     */
    {					/* make sure it's on top	  */
      wind_set (win_handle, WF_TOP, 0L, 0L);
      return (win_handle);              /* and return the handle.         */
    }

  /* center initial window tree (dialog), sizes to trect */
  frmx_center(tree_wind, &trect);

  /* calc window fullsize */
  winx_calc (WC_BORDER, WIN_KIND, trect, &wrect);

  win_handle = wind_create (WIN_KIND, wrect);            /* create window */

  if (win_handle < 0)
    {
      form_alert (1, ALERT_WIN);
      win_handle = NO_WINDOW;
    }
  else
    {
      wind_set (win_handle, WF_NAME, win_name, 0L); /* set window title   */
      wind_open (win_handle, wrect);                /* open window        */
    }

    winx_get (win_handle, WF_WORKXYWH, &win_rect);

    return (win_handle);
}

/* Close and delete a window. */
void w_close (void)
{
  /* only do close processing */
  if (win_handle != NO_WINDOW)
    {
      /* if the window is open! */
      wind_close (win_handle);
      wind_delete (win_handle);
      win_handle = NO_WINDOW;
    }
}

/* ------------------------------------
 * Standard AES redraw handler for trees in windows (non-modal dialogs).
 */
void do_redraw (short window, OBJECT *ptree, GRECT *predrawrect)
{
  short  calltype;
  short  doneflag;
  GRECT  t1;

  /* process the rectangle list for the window.  for each intersection of
   * the area to be redrawn with a visible window rectangle, call objc_draw
   * to draw that portion of the menu.
   */
  wind_update (BEG_UPDATE);

  calltype = WF_FIRSTXYWH;
  doneflag = FALSE;

  do
    {
      winx_get (window, calltype, &t1);
      /* something to draw? */
      if (t1.g_w && t1.g_h)
        {
          if (rc_intersect (predrawrect, &t1) && t1.g_w && t1.g_h)
              objc_draw (ptree, R_TREE, MAX_DEPTH, t1);
        }
      else
        doneflag = TRUE;

      calltype = WF_NEXTXYWH;
    }
  while (doneflag == FALSE);

  wind_update (END_UPDATE);
}

/* Standard message handling code. */
void msg_handler (short *msgbuf)
{
  OBJECT  *ptree = tree_wind;
  GRECT  tmp;

  switch (msgbuf[0])
    {
    case AC_OPEN:
      w_open ();
      break;

    case AC_CLOSE:
    case WM_CLOSED:
    case AP_TERM:
      prg_exit ();
      break;

    case WM_TOPPED:
    case WM_NEWTOP:
      wind_set (msgbuf[3], WF_TOP, 0L, 0L);
      break;

    case WM_UNTOPPED:		/* aes 3.1 */
    case WM_BOTTOMED:		/* winx    */
      wind_set (msgbuf[3], WF_BOTTOM, 0L, 0L);
      break;

    case WM_REDRAW:
      do_redraw (msgbuf[3], ptree, (GRECT *) &msgbuf[4]);
      break;

    case WM_MOVED:
    case WM_SIZED:
      /* get new window place */
      tmp = *(GRECT *)(&msgbuf[4]);
      winx_calc (WC_WORK, WIN_KIND, tmp, &win_rect);

      /* set new dialog place */
      ptree->ob_x = win_rect.g_x;
      ptree->ob_y = win_rect.g_y;

      /* set new window place */
      winx_calc (WC_BORDER, WIN_KIND, win_rect, &tmp);
      wind_set (msgbuf[3], WF_CURRXYWH, tmp);

      break;

    } /* END switch (message) */
}

/* ------------------------------------
 * Handle a button click in a dialog window.
 *
 * This routine handles radio button clicks, and clicks on SELECTABLE
 * objects that are not DISABLED.  It basically mimics the actions of
 * the form_do() click handler...radio button clicks are processed
 * immediately, then the mouse is ignored until the button is released.
 * Clicks on non-radio objects will track the mouse, and will handle
 * the object based on where the mouse is when the button is released.
 */
int mouse_handler (int mx, int my, int mb)
{
  int     dummy;
  int     selobj;
  int     in_obj = FALSE;
  OBJECT *ptree = tree_wind;
  OBJECT *pobj;

  selobj = objc_find (ptree, R_TREE, MAX_DEPTH, mx, my);
  pobj   = &ptree[selobj];

  if (selobj <= R_TREE || (pobj->ob_state & DISABLED))
    return NO_OBJECT;

  wind_update (BEG_UPDATE);

  if (pobj->ob_flags & RBUTTON)
    {
      /* handle radio buttons */
      rb_select (ptree, selobj);

      if ((pobj->ob_flags & TOUCHEXIT) == FALSE)
        evnt_button (1,1,0, &dummy, &dummy, &dummy, &dummy);

      in_obj = TRUE;
    }
  else
    {
      /* handle non-radio buttons */
      if (pobj->ob_flags & TOUCHEXIT)
        {
          if (pobj->ob_flags & SELECTABLE)
            {
              pobj->ob_state ^= SELECTED;
              objc_draw (ptree, selobj, MAX_DEPTH, win_rect);
            }
          in_obj = TRUE;
        }
      else
        {
          /* not a touchexit object... */
          if (pobj->ob_flags & SELECTABLE)
            {
              in_obj = graf_watchbox (ptree, selobj,
                              (pobj->ob_state ^ SELECTED),
                               pobj->ob_state);
            }
        }
    }
  wind_update (END_UPDATE);

  if (in_obj == FALSE)
    selobj = NO_OBJECT;             /* Nothing changed          */
  else
    {
      if ((pobj->ob_flags & (TOUCHEXIT|EXIT)) == FALSE)
        selobj = 0;                 /* Click on non-EXIT object */
    }
  return (selobj);
}

/* check if key is a dialog shortcut and select the object indicated */
int key_handler (char key)
{
  int  selobj;

  /* check if key is a shortcut and select it */
  selobj = key_to_obj (tree_wind, key);
  if (selobj != NO_OBJECT &&
    (tree_wind[selobj].ob_state & DISABLED) == FALSE)
    {
      /* a radio button? */
      if ((tree_wind[selobj].ob_flags & RBUTTON))
        rb_select (tree_wind, selobj);
      else
        {
          /* a normal selectable object? */
          if (tree_wind[selobj].ob_flags & SELECTABLE)
            {
              /* select object */
              tree_wind[selobj].ob_state ^= SELECTED;
              objc_draw (tree_wind, selobj, MAX_DEPTH, win_rect);
            }
        }
    }
  return (selobj);
}

/* Visually select indicated radio button, de-sel old button. */
void rb_select (OBJECT *ptree, int selobj)
{
  int  oldobj;

  oldobj = obj_rbselect (ptree, selobj, SELECTED);
  objc_draw (ptree, selobj, MAX_DEPTH, win_rect);

  if (oldobj > R_TREE && oldobj != selobj)
    objc_draw (ptree, oldobj, MAX_DEPTH, win_rect);
}

/* ------------------------------------
 * Attempt to translate a keystroke into an object.
 *
 * This routine scans every object in a tree, comparing each ob_state's
 * hi-byte value the key value passed to the routine.  If a match
 * is found, the object number is returned, else NO_OBJECT is returned.
 */
int key_to_obj (OBJECT *ptree, char key)
{
  int  curobj;

  curobj = -1;
  do
    {
      curobj ++;
      /* Compare key to the extended object state (bits 8-15) */
      if (key == ((ptree->ob_state >> 8) & 0x00FF))
        return (curobj);
    }
  while (((ptree++)->ob_flags & LASTOB) == FALSE);

  return (NO_OBJECT);
}

void de_select (OBJECT *ptree, int selobj)
{
  OBJECT *pobj;

  pobj = &ptree[selobj];

  if (!(pobj->ob_flags & RBUTTON))
    {
      if (pobj->ob_flags & SELECTABLE)
        {
          wind_update (BEG_UPDATE);
          pobj->ob_state &= (0xFFFF ^ SELECTED);
          objc_draw (ptree, selobj, MAX_DEPTH, win_rect);
          wind_update (END_UPDATE);
        }
    }
}

/* Popup - menu handler (this is all mine!!! ;-)
 * popup = address of the popup dialog & number of 'default object'.
 * parent & selobj = from which object popup menu is to be 'hanged'.
 * NOTICE: popup.defobj is aligned with parent.selobj!
 * Popup may contain any objects (text, images etc.)
 * Returns selected object number or NO_OBJECT.
 *
 * The 'popup box' should not have a border as that will leave some
 * traces to screen when redrawn.
 */
int popup (OBJECT *popup, int defobj, OBJECT *parent, int selobj)
{
  OBJECT *item;
  short  x, y, w, h, status;
  short  evnt_m, pmx, pmy, check = 1, i, m_state = 0;
  short  mbuf[8], events = (MU_BUTTON | MU_M1);

  /* popup size & place */
  form_center(popup, &x, &y, &w, &h);
  objc_offset (parent, selobj, &x, &y);

  /* NO ONBJECT -> popup below & fit into desktop window */
  if (defobj == -1)
  {
    defobj = 0;
    y += parent[selobj].ob_height;
    x = min(max(gl_rwdesk.g_x, x), gl_rwdesk.g_x + gl_rwdesk.g_w - w);
    y = min(max(gl_rwdesk.g_y, y), gl_rwdesk.g_y + gl_rwdesk.g_h - h);
  }
  else
  {
    x = min(max(gl_rwdesk.g_x, x) - popup[defobj].ob_x, gl_rwdesk.g_x + gl_rwdesk.g_w - w);
    y = min(max(gl_rwdesk.g_y, y) - popup[defobj].ob_y, gl_rwdesk.g_y + gl_rwdesk.g_h - h);
  }
  popup->ob_x = x;
  popup->ob_y = y;

  /* select current popup item */
  item = &popup[defobj];
  if (defobj && (item->ob_flags & SELECTABLE) && !(item->ob_state & DISABLED))
    item->ob_state ^= SELECTED;

  /* draw popup */
  wind_update (BEG_UPDATE);
  form_dial (0, x, y, w, h, x, y, w, h);
  objc_draw (popup, R_TREE, MAX_DEPTH, gl_rwdesk);

  status = item->ob_state;
  do
    {
      /* wait until mouse button released or mouse goes outside popup item */
      evnt_m = evnt_multi(events, 1, 1, m_state,
               check, x + item->ob_x, y + item->ob_y, item->ob_width, item->ob_height,
               0, 0, 0, 0, 0, mbuf, 0, 0,
               &pmx, &pmy, &i ,&i, &i, &i);


      /* redraw old item */
      if (check)
        {
          /* deselect old item */
          if (defobj && (item->ob_flags & SELECTABLE) && !(item->ob_state & DISABLED))
            {
              status ^= SELECTED;
              objc_change (popup, defobj, 0, x + item->ob_x, y + item->ob_y, item->ob_width, item->ob_height, status, 1);
            }
        }
      else
        {
          check = 1;
          x = popup->ob_x;
          y = popup->ob_y;
        }

      /* checking for MU_M1 might give some trouble with doubleclicks */
      if (!(evnt_m & MU_BUTTON))
        {
          /* find new object */
          defobj = objc_find (popup, R_TREE, MAX_DEPTH, pmx, pmy);
          if (defobj != NO_OBJECT)
            {
              /* select new item */
              item = &popup[defobj];
              status = item->ob_state;
              /* only show if intended */
              if (defobj && (item->ob_flags & SELECTABLE) && !(item->ob_state & DISABLED))
                {
                  status ^= SELECTED;
                  objc_change (popup, defobj, 0, x + item->ob_x, y + item->ob_y, item->ob_width, item->ob_height, status, 1);
                }
            }
          /* outside of the whole popup menu */
          else
            {
              check = x = y = 0;
              item = popup;
            }
        }
      else
        {
          /* if mouse button released on 'parent' button, wait for new click */
          if (m_state == 0 && objc_find (parent, R_TREE, MAX_DEPTH, pmx, pmy) == selobj)
            {
              /* change click and drag into ST-style click and click */
              if (defobj && (item->ob_flags & SELECTABLE) && !(item->ob_state & DISABLED))
                {
                  status ^= SELECTED;
                  objc_change (popup, defobj, 0, x + item->ob_x, y + item->ob_y, item->ob_width, item->ob_height, status, 1);
                }
              m_state = 1;
              evnt_m = 0;
            }
        }
    }
  while (!(evnt_m & MU_BUTTON));

  evnt_button(1, 1, 0, &i, &i, &i, &i);		/* until button up */

  /* restore area below popup */
  form_dial (3, x, y, w, h, x, y, w, h);
  wind_update (END_UPDATE);

  /* check if object was meant for selection */
  if (defobj != NO_OBJECT && (!(item->ob_flags & SELECTABLE) || (item->ob_state & DISABLED)))
      return (NO_OBJECT);

  return (defobj);
}

/* Display and 'execute' tree number 'rsc_tree'. Return exit object number. */
int do_dialog (int rsc_tree)
{
  OBJECT  *ptree;
  GRECT    place;
  int      i;

  rsrc_gaddr (R_TREE, rsc_tree, &ptree);
  frmx_center(ptree, &place);

  /* draw dialog */
  wind_update (BEG_UPDATE);
  form_dial (0, 0, 0, 0, 0, place.g_x, place.g_y, place.g_w, place.g_h);
  objc_draw (ptree, R_TREE, MAX_DEPTH, place);

  i = form_do (ptree, R_TREE);
  ptree[i].ob_state ^= SELECTED;

  /* restore area below popup */
  form_dial (3, 0, 0, 0, 0, place.g_x, place.g_y, place.g_w, place.g_h);
  wind_update (END_UPDATE);

  return (i);
}

/* cut a filename or -mask away from the filepath */
char *pathcut (char *path)
{
  int i = 0;

  while (path[i ++] != '\0');		/* search end of string      */
  while (path[-- i] != '\\' && i >= 0);	/* search path separator     */
  path[++ i] = '\0';			/* cut path from name / mask */

  return (path);
}

/* ------------------------------------
 * Selection handling functions.
 */
void do_menu (short entry)
{
  static short msg[8];
  int help;

  switch (entry)
    {
    case 'I':
    case POP_INFO:
      do_dialog (TREEINFO);
      break;

    case HELP:
    case POP_HELP:
      if ((help = appl_find(HELP_PRG)) == NO_OBJECT)
        form_alert (1, ALERT_HELP);
      else
        {
          /* Event message buffer for running help. */
          msg[0] = VA_START;
          msg[1] = gl_apid;
          msg[2] = 0;
          /* AV-message address into 3rd & 4th places. */
          *(char **)&msg[3] = HELPFILE;
          msg[5] = 0;
          msg[6] = 0;
          msg[7] = 0;

          /* ACC activate */
          appl_write(help, 16, msg);
        }
      break;

    case 'X':
    case 'C':
      write_clip(show.result);
      break;

    case 'V':
      read_clip();
      break;

    case 'O':
    case POP_OPEN:
      open_file (TRUE);
      break;

    case 'S':
    case POP_SAVE:
      open_file (FALSE);
      break;

    case 'D':
    case POP_VARS:
      clear_vars ();
      break;

    case 'Q':
      prg_exit ();
      break;		/* acc's don't exit... */

    /* ctrl + arrow left */
    case 0x7300:
      show.pos = show.idx = 0;
      show_string ();
      break;

    /* ctrl + arrow right */
    case 0x7400:
      show.pos = show.len;
      show.idx = max (0, show.len - SHOWLEN);
      show_string ();
      break;
    }
}

/* initialize clipboard path */
char *scrap_init(void)
{
  static int initialized = FALSE;
  static char scrap[PATHSIZE] = "";
  char *path;
  long handle, len;

  if(initialized)
    return scrap;

  scrp_read(scrap);
  /* check clipboard existence */
  if (scrap[0]=='\0')
  {
    if ((path = getenv("CLIPBOARD")) != NULL ||
        (path = getenv("CLIPBRD"))   != NULL ||
        (path = getenv("SCRAPDIR"))  != NULL)
      strcpy(scrap, path);
    else
    {
      long ssp;

      /* onto bootdevice */
      strcpy(scrap, "X:\\clipbrd");
      ssp = Super(NULL);
      *scrap = (char) (*((int *) 0x446)+65);
      Super((void *) ssp);
    }
  }

  /* if clipboard doesn't exist, create it */
  if ((len = strlen(scrap)) > 0)
  {
    len--;
    if (scrap[len]=='\\')
      scrap[len]='\0';

    handle = Dcreate(scrap);
    if (handle>=0 || handle==-36)
      strcat(scrap, "\\");
  }
  scrp_write(scrap);
  strcat(scrap, "scrap.txt");
  initialized = TRUE;
  return scrap;
}

/* read an expression from the clipboard */
void read_clip(void)
{
  char *scrap;
  FILE *fp;

  scrap = scrap_init();

  if((fp = fopen(scrap, "rb")))
  {
    fgets(show.exp, MAX_LN, fp);
    fclose(fp);

    /* strip line feed(s) */
    show.len = strlen(show.exp);
    while(--show.len >= 0 && show.exp[show.len] < ' ');
    show.exp[++show.len] = '\0';

    show.pos = show.len;
    show_string();
    evaluate();
  }
  else
    form_alert(1, ALERT_CLIP);
}

/* write the result to the clipboard */
void write_clip(char *expression)
{
  char *scrap;
  FILE *fp;

  scrap = scrap_init();

  if((fp = fopen(scrap, "wb")))
  {
    fputs(expression, fp);
    fclose(fp);
  }
  else
    form_alert(1, ALERT_SCRAP);
}

/* Menu innards...
 * clear additional variable names and all variable values
 */
void clear_vars (void)
{
  int i, idx, lenght;

  /* check that variables are where they should be... */
  if (comm.vname[32] == '\0' && comm.vname[33] == 'E' && comm.vname[34] == '\0')
    {
      if (form_alert (2, ALERT_CLEAR) == 1)
        {
          i = 35;
          idx = 12;
          while ((lenght = strlen (&comm.vname[i])) > 0)
            {
              comm.vname[i] = 0;
              comm.value[idx].real = 0.0;
              comm.value[idx].imag = 0.0;
              i += ++ lenght;
              idx ++;
            }
        }
    }
}

/* read and execute expressions */
void open_file (int mode_open)
{
  /* both path & name */
  char name[256];
  int  i, idx, lenght;
  FILE *fp;

  if (select_file (filepath, filename, (mode_open == TRUE ? FSEL_OPEN: FSEL_SAVE)))
    {
      /* whole filename */
      strcpy (name, pathcut (filepath));
      strcat (name, filename);

      if (mode_open)
        fp = fopen (name, "r");
      else
        fp = fopen (name, "w");

      if (fp != NULL)
        {
          if (mode_open)
            {
              /* read and execute expressions */
              do
                {
                  fgets (show.exp, MAX_LN, fp);
                  i = strlen (show.exp);
                  while (i >= 0 && show.exp[i] < ' ')
                    show.exp[i --] = 0;		/* remove '\n's */
                  show.len = ++ i;
                  evaluate ();
                }
              while (comm.error <= NO_EXP && !feof (fp));
            }
          else
            {
              /* save all variables */
              fprintf (fp, "; Clac variables.\n");
              i = 0; idx = 0;
              while ((lenght = strlen (&comm.vname[i])) > 0)
                {
                  conv (comm.value[idx], name, DEC, CPX_CMPNT, RAD);
                  fprintf (fp, "%s=%s\n", &comm.vname[i], name);
                  i += ++ lenght;
                  idx ++;
                }
            }
          fclose (fp);
        }
    }
}

/* ------------------------------------ */
void do_object (int object, short key, short state)
{
  /* for popup menu */
  short i;

  switch (object)
    {
  /* memory variables */
    case MEM0:
      memory (1, "F1", state);
      break;
    case MEM1:
      memory (2, "F2", state);
      break;
    case MEM2:
      memory (3, "F3", state);
      break;
    case MEM3:
      memory (4, "F4", state);
      break;
    case MEM4:
      memory (5, "F5", state);
      break;
    case MEM5:
      memory (6, "F6", state);
      break;
    case MEM6:
      memory (7, "F7", state);
      break;
    case MEM7:
      memory (8, "F8", state);
      break;
    case MEM8:
      memory (9, "F9", state);
      break;
    case MEM9:
      memory (0, "F0", state);
      break;
    case PIII:
      memory (10, "PI", 0);
      break;
  /* different modes (output, input) */
    case MODECPLX:
      change_mode (tree_cplx, &mode_cplx, MODECPLX);
      break;
    case MODETRIG:
      change_mode (tree_trig, &mode_trig, MODETRIG);
      break;
    case MODEOUT:
      change_mode (tree_mode, &mode_out, MODEOUT);
      break;
  /* menu popup */
    case HELPER:
      if ((i = popup (tree_menu, -1, tree_wind, HELPER)) != NO_OBJECT)
        do_menu (i);
      break;
  /* number and operator entering */
    case POINT:
    case ZERO:
    case ONE:
    case TWO:
    case THREE:
    case FOUR:
    case FIVE:
    case SIX:
    case SEVEN:
    case EIGHT:
    case NINE:
    case MULTIPLY:
    case DIVIDE:
    case PLUS:
    case MINUS:
    case TOPOWER:
    case PARLEFT:
    case PARIGHT:
      add_char (ex[object]);
      break;
  /* expression entering */
    case HUP:			/* history */
    case HDOWN:
      handle_history (object);
      break;
    case CLEAR:
      show.exp[0] = show.show[0] = show.idx = show.pos = show.len = 0;
      show_string ();
      break;
    case DELCHAR:
      del_char ();
      break;
    case MOVELFT:
      if (show.pos > 0)
        show.pos --;
      show_string ();
      break;
    case MOVERGHT:
      if (show.pos < show.len)
        show.pos ++;
      show_string ();
      break;
    case EVALUATE:
      evaluate ();
      break;
  /* a plain key */
    case 0:
      switch (key)
        {
          case TAB:
            /* If expression, evaluate. */
            if (show.exp[0] != 0)
              evaluate ();

            /* put result back into expression string */
            strcpy (show.exp, show.result);
            show.pos = show.len = strlen (show.exp);
            show.idx = max (0, show.len - SHOWLEN);
            show_string ();
            break;
          case DEL:
            if (show.len > show.pos)
              {
                memmove (&show.exp[show.pos], &show.exp[show.pos + 1], show.len-- - show.pos);
                show_string ();
              }
            break;
          default:
            /* if not a control char, it's probably fuction or var name */
            if (key >= ' ')
                add_char (key);
        }
    }
}

/* Dialog innards... */

/* change and output memory variables */
void memory (int i, char *var, int state)
{
  if (state & (K_LSHIFT | K_RSHIFT))
    comm.value[i] = comm.result;
  else
    {
      add_char (var[0]);
      add_char (var[1]);
      conv (comm.value[i], show.result, mode_out, mode_cplx, comm.mode_trig);
      objc_draw (tree_wind, PANEL, MAX_DEPTH, win_rect);
    }
}

/* change modes (output/input) */
void change_mode (OBJECT *ptree, int *mode, int selobj)
{
  int i;

  if ((i = popup (ptree, *mode, tree_wind, selobj)) != NO_OBJECT)
    {
      /* change complex output mode (text) */
      tree_wind[selobj].ob_spec.obspec = ptree[i].ob_spec.obspec;
      *mode = i;
    }
}

/* manage history buffer */
void handle_history (int direction)
{
  if (direction == HUP)
    cur_hist = cur_hist->prev;
  else
    cur_hist = cur_hist->next;

  strcpy (show.exp, cur_hist->exp);
  show.pos = show.len = cur_hist->explen;
  show.idx = max (0, show.len - SHOWLEN);
  show_string ();
}

/* add a char to the expression */
void add_char (char key)
{
  if (show.len < MAX_LN)
    {
      /* make room for a char */
      if (show.len > show.pos)
        memmove (&show.exp[show.pos + 1], &show.exp[show.pos], show.len - show.pos);

      /* insert char & new string end mark into expression */
      show.exp[show.pos ++] = key;
      show.exp[++ show.len] = '\0';

      show_string ();
    }
}

/* delete a char from an expression */
void del_char (void)
{
  if (show.pos > 0)
    {
      if (show.pos < show.len)
        memmove (&show.exp[show.pos - 1], &show.exp[show.pos], show.len - show.pos);

      show.exp[-- show.len] = '\0';
      show.pos --;

      show_string ();
    }
}

/* show fitting part of en expression in dialog string */
void show_string (void)
{
  /* increase show-index, if cursor would not fit into the shown string */
  if (show.idx + SHOWLEN < show.pos)
    show.idx ++;

  /* decrease show-index, if cursor would not fit into the shown string */
  if (show.idx > show.pos)
    show.idx --;

  /* copy fitting part of expression for showing */
  memcpy (show.show, &show.exp[show.idx], SHOWLEN);
  show.show[SHOWLEN]='\0';

  /* show modified part */
  objc_draw (tree_wind, EXPRESS, MAX_DEPTH, win_rect);

  /* show cursor place */
  tree_wind[TCURSOR].ob_x = tree_wind[EXPRESS].ob_x + (show.pos - show.idx) * show.charlen;
  objc_draw (tree_wind, TCURSOR, MAX_DEPTH, win_rect);
}

/* ------------------------------------
 * evaluate given expression
 */
void evaluate(void)
{
  /* variable name */
  char name[MAX_VAR + 1];
  int i = 0;

  name[0] = '\0';
  comm.error = 0;
  show.message[0] = 0;
  comm.mode_trig = mode_trig;
  comm.exp = show.exp;

  while(comm.exp[i] != ';' && comm.exp[i] >= ' ' && comm.exp[i] != '=')
    i ++;

  /* variable assignment? */
  if (comm.exp[i] == '=')
    comm.exp = var_get (i, name, comm.exp, &comm.error);
  else
    i = 0;

  /* not error in variable assignement -> CALCULATE */
  if (!comm.error)
    calc (&comm);

  /* result into a string according to output mode */
  if (!comm.error)
    comm.error = conv (comm.result, show.result, mode_out, mode_cplx, comm.mode_trig);

  /* get possible error text */
  if (comm.error)
    errors (comm.error, show.message);
  else
    {
      /* variable assignment */
      if (name[0])
        var_put (name, &comm);
      if (comm.error)
        errors (comm.error, show.message);
    }

  /* show result & possible message */
  objc_draw (tree_wind, PANEL, MAX_DEPTH, win_rect);
  objc_draw (tree_wind, MESSAGES, MAX_DEPTH, win_rect);

  if (comm.error)
    {
      if (comm.err_pos)
        {
          /* set cursor to the place of error */
          show.pos = i + comm.err_pos;
          show.idx = max (0, show.pos - SHOWLEN / 2);
        }
    }
  else
    {
      /* copy expression to history */
      strcpy (history->exp, show.exp);
      history->explen = show.len;
      history = history->next;
      cur_hist = history;

      /* clear expression */
      show.exp[0] = show.show[0] = show.idx = show.pos = show.len = 0;
    }
  show_string ();
}
